#include <mpi.h>
#include <cstdio>
#include <cassert>
#include <vector>
#include <algorithm>

using namespace std;

struct mypair {
  int a,final_pos;
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void vecperm_init(vector<int> &A, vector<int> &P,
                  int N,MPI_Comm comm) {
  int myrank, size;
  MPI_Comm_rank(comm,&myrank);
  MPI_Comm_size(comm,&size);
  vector<int> Nh(size);
  for (int j=0; j<size; j++) 
    Nh[j] = N/size+(j<N%size);

  vector<mypair> AP(N);
  // Bsol.resize(N);
  if (!myrank) {
    for (int j=0; j<N; j++) {
      AP[j].a = rand()%1000;
      AP[j].final_pos = j;
    }
    random_shuffle(AP.begin(),AP.end());
  }
  MPI_Bcast(&AP[0],N*sizeof(mypair),MPI_CHAR,0,MPI_COMM_WORLD);

  // for (int j=0; j<N; j++)
  //  Bsol[j] = AP[j].a;

  int Nhere = Nh[myrank];
  A.resize(Nhere);
  P.resize(Nhere);
  int base=0;
  for (int j=0; j<myrank; j++) 
    base += Nh[j];
  // my_displ = base;

  for (int j=0; j<Nhere; j++) {
    A[j] = AP[base+j].a;
    P[j] = AP[base+j].final_pos;
  }  
}
/* //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void vecperm_init(int **A, int **P,
                  int N,int *Nherep,MPI_Comm comm) {
  vector<int> Av,Pv;
  vecperm_init(Av,Pv,N,MPI_COMM_WORLD);
  *Nherep = Av.size();
  assert(Av.size()==Pv.size());
  *A = (int *)malloc((*Nherep) * sizeof(int));
  *P = (int *)malloc((*Nherep) * sizeof(int));
  for (int j=0; j<*Nherep; j++) {
    int kk = Av[j]; 
    (*A)[j] = kk;
    (*P)[j] = Pv[j];
  }
}
*/ //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>

int main(int argc, char **argv) {
  MPI_Init(&argc,&argv);
//  int *Ap, *Pp, Nhere;
  int myrank, size;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  vector<int> A,P;
  vecperm_init(A,P,20,MPI_COMM_WORLD);
  int H = A.size();
  for(int j=0;j<H;j++){
    printf("A[%d] = %d, P[%d] = %d\n",j,A[j],j,P[j]);
  }
  MPI_Finalize();
}
