//$Id: bisecbw.cpp,v 1.6 2004/08/01 21:47:59 mstorti Exp $
#include <ctime>
#include <cmath>
#include <cstdio>
#include <cassert>
#include <cstdlib>
#include <vector>
#include <mpi.h>

using namespace std;

int find_not_asigned(const vector<int> &v, int j) {
  int k=0, n=v.size(), m=-1;
  while (k<n) {
    if(v[k]<0) m++;
    if (m==j) break;
    k++;
  }
  assert(k<n);
  return k;
}

void gen_comb(vector< vector<int> > &combinations,
	      vector<int> current, int remain) {
  if (!remain) {
    combinations.push_back(current);
  } else {
    int pj = find_not_asigned(current,0);
    for (int k=1; k<remain; k++) {
      int pk = find_not_asigned(current,k);
      // Pairing remaining processors `j' and `k'
      vector<int> current2 = current;
      current2[pj] = pk;
      current2[pk] = pj;
      gen_comb(combinations,current2,remain-2);
    }
  }
}

int main(int argc, char **argv) {
  MPI_Init(&argc,&argv);

  int  myrank, size, ierr;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  assert(!(size%2));
  vector< vector<int> > combinations;
  vector<int> current(size,-1);
  gen_comb(combinations,current,size);
  int ncombs = combinations.size();
  if (!myrank) 
    printf("total combs %d\n",ncombs);

  srand(time(NULL));
  MPI_Status status;
  int N = 100000, M=1000, sum, totsum=0;
  vector<int> buff1(N), buff2(N,0), sums(size),
    *sbuff, *rbuff, *tmp;
  if (!myrank) {
    for (int p=0; p<size; p++) {
      for (int j=0; j<N; j++) {
	int w = rand() % M;
	buff1[j] = w;
	totsum += w;
      }
      if (p<size-1)
	MPI_Send(&buff1[0],N,MPI_INT,p+1,0,MPI_COMM_WORLD);
    } 
    printf("total sum %d\n",totsum);
  } else {
    MPI_Recv(&buff1[0],N,MPI_INT,0,0,MPI_COMM_WORLD,&status);
  }

  sbuff = &buff1;
  rbuff = &buff2;

  int jcomb=0;
  while(true) {
    sum = 0;
    for (int k=0; k<N; k++) sum += (*sbuff)[k];
    MPI_Gather(&sum,1,MPI_INT,&sums[0],1,MPI_INT,0,MPI_COMM_WORLD);
    int ts = 0;
    for (int p=0; p<size; p++) ts += sums[p];
    assert(ts==totsum);
    if (!myrank) {
      printf("check sums: ");
      for (int p=0; p<size; p++) printf("%d ",sums[p]);
      printf("\n");
    }
    if (jcomb>=ncombs) break;
    vector<int> &comb = combinations[jcomb++];
    int partner = comb[myrank];
    assert(partner!=myrank);
    if (!myrank) {
      printf("Applying combination %d\n",jcomb);
      for (int p=0; p<size; p++) printf("%d ",comb[p]);
      printf("\n");
    }

    if (partner < myrank) {
      MPI_Send(&(*sbuff)[0],N,MPI_INT,partner,jcomb,MPI_COMM_WORLD);
      MPI_Recv(&(*rbuff)[0],N,MPI_INT,partner,jcomb,MPI_COMM_WORLD, &status);
    } else {
      MPI_Recv(&(*rbuff)[0],N,MPI_INT,partner,jcomb,MPI_COMM_WORLD, &status);
      MPI_Send(&(*sbuff)[0],N,MPI_INT,partner,jcomb,MPI_COMM_WORLD);
    }
    tmp = sbuff; sbuff = rbuff; rbuff = tmp;
  }
  MPI_Finalize();
}
