//$Id: bisecbw.cpp,v 1.6 2004/08/01 21:47:59 mstorti Exp $
#include <ctime>
#include <cmath>
#include <cstdio>
#include <cassert>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <algorithm>
#include <mpi.h>

using namespace std;

int checksum(const vector<int> &v) {
  int cs = 0, N = v.size();
  for (int j=0; j<N; j++) cs += v[j];
  return cs;
}

int main(int argc, char **argv) {
  MPI_Init(&argc,&argv);

  int  myrank, size;
  MPI_Status status;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  assert((size%2)==0);

  srand(myrank);
  int N=10000, ntimes=10000;
  vector<int> indx(size), partner(size), buff1(N), buff2(N,0), cs(size);
  for (int j=0; j<N; j++) buff1[j] = rand();
  int mycs = checksum(buff1);
  MPI_Allgather(&mycs,1,MPI_INT,
                &cs[0],1,MPI_INT, MPI_COMM_WORLD);
  //#define DOCS
#ifdef DOCS
  if (!myrank) {
    printf("Before scatter: ---------\n");
    for (int j=0; j<size; j++ ) 
      printf("proc %d, checksum %x\n", j,cs[j]);
  }
#endif

  // printf("[%d] checksum %x\n",myrank,);
  if (!myrank) {
    for (int j=0; j<size; j++) indx[j] = j;
    random_shuffle(indx.begin(),indx.end());
    for (int j=0; j<size/2; j++) {
      int j1= indx[2*j];
      int j2= indx[2*j+1];
      partner[j1] = j2;
      partner[j2] = j1;
    }
  }
  MPI_Bcast(&partner[0],size,MPI_INT,0,MPI_COMM_WORLD);
  int mypartner = partner[myrank];

#ifdef DOCS
  printf("I'm %d, my partner is %d\n",myrank,mypartner);
#endif

  MPI_Barrier(MPI_COMM_WORLD);
  double start = MPI_Wtime();
  for (int j=0; j<ntimes; j++) {
    if (myrank<mypartner) {
      MPI_Send(&buff1[0],N,MPI_INT,mypartner,0,MPI_COMM_WORLD);
      MPI_Recv(&buff1[0],N,MPI_INT,mypartner,0,MPI_COMM_WORLD, &status);
    } else {
      MPI_Recv(&buff2[0],N,MPI_INT,mypartner,0,MPI_COMM_WORLD, &status);
      MPI_Send(&buff1[0],N,MPI_INT,mypartner,0,MPI_COMM_WORLD);
      memcpy(&buff1[0],&buff2[0],N*sizeof(int));
    }
  }
  MPI_Barrier(MPI_COMM_WORLD);
  double elapsed = MPI_Wtime()-start;
#define D double
  double bw = D(N)*D(size)*D(sizeof(int)*8)*D(ntimes)/elapsed/1.0e9;
  if (!myrank) printf("elapsed %fsecs, bw %fGbit/sec\n",elapsed,bw);

  // printf("[%d] checksum %x\n",myrank,checksum(buff1));
  mycs = checksum(buff1);
  MPI_Allgather(&mycs,1,MPI_INT,
                &cs[0],1,MPI_INT, MPI_COMM_WORLD);
#ifdef DOCS
  if (!myrank) {
    printf("After scatter: ---------\n");
    for (int j=0; j<size; j++ ) 
      printf("proc %d, checksum %x\n", j,cs[j]);
  }
#endif

  MPI_Finalize();
}
