// $Id: decomp.cpp,v 1.3 2004/10/03 19:52:18 mstorti Exp $
#include <vector>
#include <cstdio>
#include <cmath>

using namespace std;

void decomp(int n,vector<double> &weights, //L-decomp-b
	    vector<int> &nrows) { 
  int size = weights.size();
  nrows.resize(size);
  double sum_w = 0., carry=0., a;
  for (int p=0; p<size; p++) sum_w += weights[p];
  int j = 0;
  double tol = 1e-8;
  for (int p=0; p<size; p++) {
    double w = weights[p]/sum_w;
    a = w*n + carry + tol;
    nrows[p] = int(floor(a));
    carry = a - nrows[p] - tol;
    j += nrows[p];
  }
  assert(j==n);
  assert(fabs(carry) < tol);
}               
                                  //L-decomp-e
int main() {
  int NP = 10, n=1000;
  vector<double> weights(NP);
  vector<int> nrows;
  double sumw=0.;
  for (int j=0; j<NP; j++) {
    weights[j] = double(rand())/double(RAND_MAX);
    sumw += weights[j];
  }
  for (int j=0; j<NP; j++) weights[j] /= sumw;

  decomp(n,weights,nrows);
  int nrowtot = 0;
  for (int j=0; j<NP; j++) {
    printf("in [%d] weight %f, wants %f, assigned %d\n",
	   j,weights[j],weights[j]*n,nrows[j]);
    nrowtot += nrows[j];
  }
  printf("total rows %d\n",nrowtot);
}
