// $Id: poisson.cpp,v 1.3 2004/10/03 19:52:18 mstorti Exp $
#include <mpi.h>
#include <mpe.h>
#include <cstdio>

void exchng1(double *a,int n,int s,int e, //L-ex-b
	     MPI_Comm comm1d,int nbrbot,int nbrtop) {
  MPI_Status status;
#define ROW(j) (&a[(j-s+1)*(n+1)])

  // Exchange top row
  MPI_Send(ROW(e-1),n+1,MPI_DOUBLE,nbrtop,0,comm1d);
  MPI_Recv(ROW(s-1),n+1,MPI_DOUBLE,nbrbot,0,comm1d,&status);

  // Exchange top row
  MPI_Send(ROW(s),n+1,MPI_DOUBLE,nbrbot,0,comm1d);
  MPI_Recv(ROW(e),n+1,MPI_DOUBLE,nbrtop,0,comm1d,&status);
}
//L-ex-e

void sweep(double *u1,double *u2,double *f,  //L-tools-b
	   int n,int s,int e) { } 

double diff(double *u1,double *u2, 
	    int n,int s,int e) { }
//L-tools-e

int main() { //L-main-b
  int n=10;
  
  // Get MPI info
  int size,rank;
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  int periods = 0;
  MPI_Comm comm1d;
  MPI_Cart_create(MPI_COMM_WORLD,1,&size,&periods,1,&comm1d);
  MPI_Comm_rank(MPI_COMM_WORLD,&rank);
  
  int direction=1, nbrbot, nbrtop;
  MPI_Cart_shift(comm1d,direction,1,&nbrbot,&nbrtop);

  int s,e;
  MPE_Decomp1d(n-1,size,rank,&s,&e);

  int 
    rows_here = e-s+2,
    points_here = rows_here*(n+1);
  double 
    *tmp,
    *u1 = new double[points_here],
    *u2 = new double[points_here],
    *f = new double[points_here];
#define U1(i,j) u1((n+1)*(j-s+1)+i)
#define U2(i,j) u2((n+1)*(j-s+1)+i)
#define F(i,j) f((n+1)*(j-s+1)+i)

  // Assign bdry conditions to 'u1' and 'u2' ...
  // Define 'f' ...
  // Initialize 'u1' and 'u2'

  // Do computations
  int iter, itmax = 1000;
  for (iter=0; iter<itmax; iter++) {
    // Communicate ghost values
    exchng1(u1,n,s,e,comm1d,nbrbot,nbrtop);
    // Compute new 'u' values from Jacobi iteration
    sweep(u1,u2,f,n,s,e);

    // Compute difference between
    // 'u1' (u^k) and 'u2' (u^{k+1})
    double diff_here2 = diff(u1,u2,n,s,e);
    double diff;
    MPI_Allreduce(&diff_here2,&diff,1,MPI_DOUBLE,
		  MPI_SUM,MPI_COMM_WORLD);
    diff = sqrt(diff);
    printf("iter %d, diff %d\n",iter,diff);

    // Swap pointers
    tmp=u1; u1=u2; u2=tmp;

    if (diff<1e-5) break;
  }
  
  printf("Converged in %d iterations\n",iter);

  delete[] u1;
  delete[] u2;
  delete[] f;
}
//L-main-e
