//$Id: primes2.cpp,v 1.1 2004/07/23 01:33:27 mstorti Exp $
#include <cstdio>
#include <cmath>
#include <cassert>
#include <cstdlib>
#include <ctime>
#include <unistd.h>

#include <mpi.h>

int is_prime(int n) {
  if (n<2) return 0;
  int m = int(sqrt(n));
  for (int j=2; j<=m; j++) 
    if (!(n % j)) return 0;
  return 1;
}

int main(int argc, char **argv) {
  MPI_Init(&argc,&argv);

  int  myrank, size;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  srand(time(NULL));
  int n2, primesh=0, primes, chunk,
    n1=0, max=5000000, chunkh, n1h,n2h;
  chunk = 10000+(rand()%10000);
  MPI_Bcast(&chunk,1,MPI_INT,0,MPI_COMM_WORLD);
  chunkh = chunk/size;
  assert(chunkh>0);
  chunk = chunkh*size;
  double start = MPI_Wtime();
  while(1) {
    if (n1>=max) break;
    n2 = n1 + chunk;
    if (n2>max) n2=max;
    n1h = n1+myrank*chunkh;
    n2h = n1h + chunkh;
    if (n2h>n2) n2h=n2;
    // printf("n1h %05d, n2h %05d, n1 %5d, n2 %5d, runs in [%d] primesh %d\n",
    //        n1h,n2h,n1,n2,myrank,primesh);
    for (int n=n1h; n<n2h; n++) {
      if (is_prime(n)) primesh++;
    }
    MPI_Reduce(&primesh,&primes,1,MPI_INT,
	       MPI_SUM,0,MPI_COMM_WORLD);
    if (!myrank) printf("pi(%d) = %d (elapsed %.4f)\n",
			n2,primes,MPI_Wtime()-start);
    n1 = n2;
  }
  MPI_Barrier(MPI_COMM_WORLD);
  
#define MAXLEN 1000
  char host[MAXLEN];
  int ierr = gethostname(host,MAXLEN);
  assert(ierr==0);
  printf("Hello world. I am %d out of %d. Runs on host %s\n",myrank,size,host);

  MPI_Finalize();
}
