void re_scatter(vector<int> &buff,int size, 
		int myrank,proc_fun proc,void *data) {
  vector<int> send_counts(size,0);
  // Check how many elements should go to
  // each processor
  int N = buff.size();
  for (int j=0; j<N; j++) {
    int p = proc(buff[j],size,data);
    send_counts[p]++;
  }

  // Dimension buffers and ptr vectors
  vector<int> send_buff(N);
  vector<int> send_ptr(size);
  vector<int> recv_counts(size);
  vector<int> recv_ptr(size);
  send_ptr[0] = 0;
  for (int j=1; j<size; j++) 
    send_ptr[j] = send_ptr[j-1] + send_counts[j-1];

  for (int j=0; j<N; j++) {
    int p = proc(buff[j],size,data);
    int pos = send_ptr[p];
    send_buff[pos] = buff[j];
    send_ptr[p]++;
  }
#if 0
  for (int j=0; j<size; j++)
    printf("at [%d], sends %d to %d\n",myrank,send_counts[j],j);
#endif
  MPI_Alltoall(&send_counts[0],1,MPI_INT,
	       &recv_counts[0],1,MPI_INT,MPI_COMM_WORLD);

#if 0
  for (int j=0; j<size; j++)
    printf("at [%d], receives %d from %d\n",myrank,recv_counts[j],j);
#endif

  recv_ptr[0] = 0;
  send_ptr[0] = 0;
  for (int j=1; j<size; j++) {
    recv_ptr[j] = recv_ptr[j-1] + recv_counts[j-1];
    send_ptr[j] = send_ptr[j-1] + send_counts[j-1];
  }

  int recv_buff_size = 0;
  for (int j=0; j<size; j++) recv_buff_size += recv_counts[j];

  buff.resize(recv_buff_size);
  MPI_Alltoallv(&send_buff[0],&send_counts[0],&send_ptr[0],MPI_INT,
		&buff[0],&recv_counts[0],&recv_ptr[0],MPI_INT,
		MPI_COMM_WORLD);
}

int main(int argc, char **argv) {
  
  MPI_Init(&argc,&argv);

  int  myrank, size, ierr;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  // srand(time(NULL));

  int N = 10, M=1000;

#if 0
  int rest = N % size;
  int Nhere = N/size + (myrank<rest ? 1 : 0);
  vector<int> buff(Nhere);
  for (int j=0; j<Nhere; j++) 
    buff[j] = rand() % M;
#else
  vector<int> buff;
  if (!myrank) {
    buff.resize(N);
    for (int j=0; j<N; j++) 
      buff[j] = rand() % M;
  }
#endif

  int s;
  print_par(buff,"initial:\n");

  for (int k=0; k<2*size; k++) {
    void *data = &k;
    proc_fun f;
#if 0
    f = key_rotate;
#elif 0
    f = scramble;
#elif 0
    f = allproc;
#elif 1
    f = rotate_elems;
#elif 0
    f = random_scatter;
    vector<int> random_key(size);
    if (!myrank) {
      for (int j=0; j<size; j++) {
	int proc = rand() % size;
	random_key[j] = proc;
	if (!proc) printf("proc %d scattered to 0\n",j);
      }
    }
    MPI_Bcast(&random_key[0],size,MPI_INT,0,MPI_COMM_WORLD);
    data = &random_key;
#else
#error Not defined scatter function
#endif

    re_scatter(buff,size,myrank,f,data);
    print_par(buff,"after rescatter: -------\n");
  }
  
  MPI_Finalize();  
}
