// $Id: rich.cpp,v 1.2 2004/10/24 21:19:15 mstorti Exp $
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
//  Solves Ax=b using Richardson
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
#include <cstdio>
#include <cassert>
#include <cmath>
#include <cfloat>
#include <cstring>
#include <mpi.h>

using namespace std;

void read_matrix(char *file,
		 int ncmax,int *I,int *J,double * AIJ,int &nc,
		 int nmax,double *B,double *X,int &n);

int main(int argc, char **argv) {

  int  myrank, size, ierr;

  MPI_Init(&argc,&argv);

  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

#define NCMAX 1000
#define NMAX 1000
  int I[NCMAX], J[NCMAX], nc,n;
  double AIJ[NCMAX],B[NMAX],X[NMAX];
  
  read_matrix("rich.dat",
	      NCMAX,I,J,AIJ,nc, 
	      NMAX,B,X,n);

  if (!myrank) {
    for (int j=0; j<n; j++) 
      printf("B(%d) = %f\n",j,B[j]);
    
    for (int j=0; j<nc; j++) 
      printf("A(%d,%d) = %f\n",I[j],J[j],AIJ[j]);
  }

  int *sendcnts = new int[size];
  int *sdispls = new int[size];
  int *recvcnts = new int[size];
  int *rdispls = new int[size];

  int rest = n % size;
  for (int j=0; j<size; j++) {
    recvcnts[j] = n/size + (j<rest ? 1 : 0);
  }
  rdispls[0] = 0;
  for (int j=1; j<size; j++) {
    rdispls[j] = rdispls[j-1] + recvcnts[j-1];
  }

  double *x = new double[n];
  double *rloc = new double[n];
  double *r = new double[n];
  for (int j=0; j<n; j++) x[j] = 0.;

  int start = rdispls[myrank];
  int end = start+recvcnts[myrank];

  double omega = 0.2, tol=1e-2, norm, norm0;
  int itmax = 100, iter;
  for (iter=0; iter<itmax; iter++) {
    for (int j=0; j<n; j++) rloc[j] = 0.;
    for (int j=0; j<nc; j++) {
    
      int eq = I[j];
      if (I[j]>=start && I[j]<end) {
	rloc[eq] += -AIJ[j]*x[J[j]];
      }
    }
    for (int j=0; j<n; j++) rloc[j] += B[j];
    MPI_Allgatherv(&rloc[start],recvcnts[myrank],MPI_DOUBLE,
		   r,recvcnts,rdispls,MPI_DOUBLE,MPI_COMM_WORLD);

#if 0
    double rnorm=0;
    for (int j=0; j<n; j++) rnorm += r[j]*r[j];
    rnorm = sqrt(rnorm);
    if (!myrank) printf("iter %d, rnorm %g\n",iter,rnorm);
    if (rnorm<tol) break;
#endif
 
#if 0
    if (!myrank) 
      for (int j=0; j<n; j++) printf("%d %f\n",j,r[j]);
#endif

    norm = 0.;
    for (int j=0; j<n; j++) {
      norm += r[j]*r[j];
      x[j] += omega*r[j];
    }
    norm = sqrt(norm);
    if (iter==0) norm0=norm;
    if (!myrank) 
      printf("iter %d, norm %e\n",iter,norm);
    if (norm<tol*norm0) break;
  }
    
  if (!myrank) {
    printf("iters %d, norm %e, solution: \n",iter,norm);
    for (int j=0; j<n; j++) 
      printf("%d %g [%g]\n",j,x[j],X[j]);
  }

  delete[] x;
  delete[] rloc;
  delete[] r;

  delete[] sendcnts;
  delete[] sdispls;
  delete[] recvcnts;
  delete[] rdispls;
  MPI_Finalize();
}
