// $Id: rich.cpp,v 1.2 2004/10/24 21:19:15 mstorti Exp $
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
//  Solves Ax=b using Richardson
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
#include <cstdio>
#include <cassert>
#include <cmath>
#include <cfloat>
#include <cstring>
#include <vector>
#include <omp.h>

using namespace std;

int modulo(int k, int n, int *div=NULL) {
  int m = k % n;
  int d = k / n;
  if (m<0) {
    m += n;
    d -= 1;
  }
  if (div) *div = d;
  return m;
}

void mat_vec(vector<int> &Jindx,vector<int > &J,
             vector<double> &AIJ, vector<double> &X, 
             vector<double> &Y) {
  int N = Jindx.size();
  int nc = AIJ.size();
#pragma omp parallel for schedule(dynamic,10000)
  for (int i=0; i<N; i++) {
    int j1 = Jindx[i];
    int j2 = (i<N-1? Jindx[i+1] : nc);
    double Yi = 0.0;
    for (int j=j1; j<j2; j++) {
      Yi += AIJ[j] * X[J[j]];
    }
    Y[i] = Yi;
  }
}

int layers=2;

void gen_matrix(int N,vector<int> &Jindx,vector<int > &J,
                vector<double> &AIJ, vector<double> &X, 
                vector<double> &B) {
  double sigma = 0.5;
  int nngbr = 2*layers+1;
  Jindx.resize(N);
  J.resize(nngbr*N);
  AIJ.resize(nngbr*N);
  X.resize(N);
  B.resize(N);

  int indx=0;
  for (int i=0; i<N; i++) {
    X[i] = i<N/2;
    Jindx[i] = indx;
    for (int j=-layers; j<=layers; j++) {
      assert(indx<J.size());
      J[indx] = modulo(i-j,N);
      AIJ[indx] = (j? -1.0/layers : 2.0+sigma);
      indx++;
    }
  }
  assert(indx==J.size());
  mat_vec(Jindx,J,AIJ,X,B);
}

int main(int argc, char **argv) {

  
  vector<int> Jindx,J;
  vector<double> AIJ, Xex, X, B, Y;
  int N=100000, itmax = 100;

  gen_matrix(N,Jindx,J,AIJ,Xex,B);
  int nc = AIJ.size();

#if 0
  for (int i=0; i<N; i++) {
    printf("i %d, B %f, X %f, A ",i,B[i],Xex[i]);
    int j1 = Jindx[i];
    int j2 = (i<N-1? Jindx[i+1] : nc);
    for (int j=j1; j<j2; j++) 
      printf(" (%d,%f)",J[j],AIJ[J[j]]);
    printf("\n");
  }
#endif

  double omega = 0.4, tol=1e-2, norm, norm0;
  int iter;
  Y.resize(N);
  X.resize(N,0.0);
  double start = omp_get_wtime();
  double rnorm;
  for (iter=0; iter<itmax; iter++) {
    rnorm=0.0;
    mat_vec(Jindx,J,AIJ,X,Y);
#pragma omp parallel for schedule(guided) reduction(+:rnorm)
    for (int i=0; i<N; i++) {
      double Ri = B[i] - Y[i];
      rnorm += Ri*Ri;
      X[i] += omega*Ri;
    }
    rnorm = sqrt(rnorm);
    // printf("iters %d, norm %g\n",iter-1,rnorm);
  }
  double erro=0.0, ee;
  for (int j=0; j<N; j++) {
    ee = X[j]-Xex[j];
    erro += ee*ee;
  }
  erro = sqrt(erro);
  printf("iter %d, norm %g, erro %g\n",iter,rnorm,erro);
  double elapsed = omp_get_wtime()-start;
  printf("elapsed %f.2secs, rate %g Gflop/sec\n",elapsed,
         2.0*(2.0*layers+1)*double(nc)*double(iter)/elapsed*1e-9);
  return 0;
}
