//$Id: scatter.cpp,v 1.4 2004/08/01 17:45:40 mstorti Exp $
#include <mpi.h>
#include <cstdio>
#include <cassert>
#include <vector>

using namespace std;

int main(int argc, char **argv) {
  MPI_Init(&argc,&argv);

  int  myrank, size;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  int N = 100; // Nbr of elements to be generated
  int M = 100000; // Max random number 
  vector<int> nelp(size,0), vhere, v, buff;
  vector<int> indx(size,0);
  if (!myrank) {
    vector<int> sum(size,0);
    v.resize(N);
    buff.resize(N); 

    for (int j=0; j<N; j++) {
      int w = rand() % M;
      int proc = w % size;
      nelp[proc]++;
      printf("w %d, -> proc%d\n",w,proc);
      v[j] = w;
      sum[proc] += w;
    }
    
    for (int p=0; p<size; p++) 
      printf("To send to [%d]: %d elements, sum %d\n",p,nelp[p],sum[p]);

    for (int p=1; p<size; p++) {
      indx[p] = indx[p-1]+nelp[p-1];
    }

    // Prepare buffer
    for (int j=0; j<N; j++) {
      int w = v[j];
      int proc = w % size;
      int k = indx[proc]++;
      buff[k] = w;
    }
    v.clear();

//     printf("Buffer to be sent\n");
//     for (int j=0; j<N; j++) 
//       printf("buff[%d] = %d\n",j,buff[j]);
  }

  MPI_Bcast(&nelp[0],size,MPI_INT,0,MPI_COMM_WORLD);
  int nhere = nelp[myrank];

  // Compute displacements
  indx[0]=0;
  for (int p=1; p<size; p++) 
    indx[p] = indx[p-1]+nelp[p-1];

  vhere.resize(nhere,-10);
  MPI_Scatterv(&buff[0],&nelp[0],&indx[0],MPI_INT,
	       &vhere[0],nhere,MPI_INT,
	       0,MPI_COMM_WORLD);

  // Check sums
  int sum = 0;
  for (int j=0; j<nhere; j++) sum += vhere[j];

  printf("[%d] %d elements, sum %d\n",myrank,nhere,sum);
  MPI_Finalize();
}
