//$Id: sqscat.cpp,v 1.5 2006/05/22 01:15:15 mstorti Exp $
#include <cstdio>
#include <cmath>
#include <cassert>
#include <cstdlib>

#include <mpi.h>

// Partitions `N' rows in `m' processes.
// Returns the start of range for processor
// `K' out of `m'. 
int range_start(int N,int m,int K) {
  int rest = N % m;
  return K*(N/m) + (K<rest? K : rest);
}

// Scatters a matrix `A (N*N)' to the other processes.
// The number of processes should be a perfect square. 
void sqscat(double *A,int N,double **B,
	    int root,int *r1,int *r2,int *c1,int *c2) {
  int  myrank, numprocs;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
  int m = int(sqrt(numprocs));

  if (numprocs!=m*m) {
    if (!myrank) 
      printf("numprocs = %d: should be a perfect square\n",
	     numprocs);
    MPI_Finalize();
    exit(0);
  }

  int rsize;
  double *sbuff;
  int *sdispls, *scnts;

  if (myrank==root) {

    sbuff = new double[N*N];
    sdispls = new int[numprocs+1];
    scnts = new int[numprocs];
    
    int proc = 0,r1,r2,c1,c2, pos=0;
    for (int I=0; I<m; I++) {
      for (int J=0; J<m; J++) {
	sdispls[proc] = pos;
	r1 = range_start(N,m,I);
	r2 = range_start(N,m,I+1);
	c1 = range_start(N,m,J);
	c2 = range_start(N,m,J+1);
	scnts[proc] = (r2-r1)*(c2-c1);
	for (int row=r1; row<r2; row++)
	  for (int col=c1; col<c2; col++)
	    sbuff[pos++] = A[row*N+col];
	proc++;
      }
    }
    
    sdispls[0] = 0;
    for (int j=1; j<numprocs; j++) 
      sdispls[j] = sdispls[j-1]+scnts[j-1];
  } 
  int 
    I = myrank/m,
    J = myrank % m;
  int proc = 0;
  *r1 = range_start(N,m,I);
  *r2 = range_start(N,m,I+1);
  *c1 = range_start(N,m,J);
  *c2 = range_start(N,m,J+1);
  rsize = (*c2-*c1)*(*r2-*r1);
  *B = new double[rsize];

  int ierr = MPI_Scatterv(sbuff,scnts,sdispls,MPI_DOUBLE,
			  *B,rsize,MPI_DOUBLE,root,MPI_COMM_WORLD);
  assert(!ierr);

  if (!myrank) {
    delete[] sbuff;
    delete[] scnts;
    delete[] sdispls;
  } 
}

int main(int argc, char **argv) {
  MPI_Init(&argc,&argv);

  int  myrank, numprocs;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&numprocs);

  const int N = 5;
  double *A, *B;
  if (!myrank) {
    A = new double[N*N];
    for (int j=0; j<N; j++) 
      for (int k=0; k<N; k++) 
	A[j*N+k] = j*100+k;
  }

  int r1,r2,c1,c2;
  sqscat(A,N,&B,0,&r1,&r2,&c1,&c2);
  printf("[%d] row range [%d,%d), col range [%d,%d)\n",
	 myrank,r1,r2,c1,c2);
  int nrows = r2-r1, ncols = c2-c1; 
  for (int j=0; j<nrows; j++) {
    printf("[%d] ",myrank);
    for (int k=0; k<ncols; k++) 
      printf("%f ",B[j*ncols+k]);
    printf("\n");
  }

  MPI_Finalize();
}
