// $Id: tsp.cpp,v 1.2 2004/10/10 14:33:14 mstorti Exp $
#include <cstdio>
#include <cassert>
#include <cmath>
#include <cfloat>
#include <cstring>
#include <mpi.h>

int MYRANK, SIZE;

using namespace std;

#define DMAX 1000.0
//#define DMAX DBL_MAX

void print_path(int *path,int Nv) {
  printf("(");
  for (int j=0; j<Nv-1; j++) 
    printf("%d,",path[j]);
  printf("%d)\n",path[Nv-1]);
}

int next_path(int *path,int N,int Np=0) {
  int j;
  for (j=0; j<N-Np; j++) if (path[j] != N-1-j) break;
  if (j==N-Np) return 0;
  else {
    if (j>0) {
      for (int k=0; k<N-j; k++) path[k] = path[k+j];
      for (int k=N-j; k<N; k++) path[k] = k;
    }
  }

  int jj=N-1-j;
  // Tries to move vertex `jj' one step to the head
  // Look where is 'jj'
  int k;
  for (k=1; k<N-1; k++) if (path[k] == jj) break;
  assert(k<N);
  // Exchange 'jj' with previous position
  path[k] = path[k-1];
  path[k-1] = jj;
  return 1;
}

double dist(int *path, int N, double *d) {
  double D=0.0;
  for (int j=0; j<N; j++) {
    int k = j+1;
    if (j==N-1) k=0;
    int J = path[j];
    int K = path[k];
#if 0
    if (!(0<=J && J<N && 0<=K && K<N)) {
      printf("[%d] J %d, K %d, j %d, k %d, Bad path ",MYRANK,J,K,j,k);
      print_path(path,N);
      assert(0);
    }
#endif
    D += d[N*J+K];
  }
  return D;
}

int main(int argc, char **argv) {
  
  MPI_Init(&argc,&argv);

  int  myrank, size, ierr;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);
  MYRANK = myrank;
  SIZE = size;
  
  int n=4, m=3;
  int N = m*n, Np=5;

  double *x = new double[2*N];
  int vertex=0;
  for (int j=0; j<m; j++) {
    double xx = double(j)/double(m-1);
    for (int k=0; k<n; k++) {
      double yy = double(k)/double(n-1);
      x[2*vertex] = xx;
      x[2*vertex+1] = yy;
      vertex++;
    }
  }

  double *d = new double[N*N];
  for (int j=0; j<N; j++) {
    for (int k=0; k<N; k++) {
      double dx = x[2*k] - x[2*j];
      double dy = x[2*k+1] - x[2*j+1];
      d[N*j+k] = sqrt(dx*dx+dy*dy);
    }
  }

  //#define DBG
  if (!myrank) {

    int *path= new int[N];
    int *pathmin = new int[N];
    int *pathslv = new int[N];
    for (int j=0; j<N; j++) path[j]=j;
    
    MPI_Status status;
    int npp=0, ncp=0;
    double dmin = DMAX;
    int dbg = 0, down=0, done=0;
    while(1) {
      if (!done) npp++;
      double dslv;
#ifdef DBG
      printf("Before recv, npp=%d\n",npp);
#endif
      MPI_Recv(&dslv,1,MPI_DOUBLE,MPI_ANY_SOURCE,MPI_ANY_TAG,
	       MPI_COMM_WORLD,&status);
      int slave = status.MPI_SOURCE;
      MPI_Recv(pathslv,N,MPI_INT,slave,MPI_ANY_TAG,
	       MPI_COMM_WORLD,&status);
#ifdef DBG
      printf("Receives from [%d] dslv %f, pathslv: ",slave,dslv);
      print_path(pathslv,N);
#endif

      if (dslv<dmin) { 
	dmin = dslv;
	memcpy(pathmin,pathslv,N*sizeof(int));
      }

      MPI_Send(path,N,MPI_INT,slave,0,MPI_COMM_WORLD);
      if (done) down++;
      if (down==size-1) break;

      if(!done && !next_path(path,Np)) {
	done = 1;
	for (int k=0; k<N; k++) path[k] = -1;
      }
    }
    printf("%d partial paths\n",npp);
    printf("min dist = %f\n",dmin);
    printf("min path = "); print_path(pathmin,N);

    delete[] path;
    delete[] pathmin;
    delete[] pathslv;
    
  } else {

    int *path= new int[N];
    int *pathmin = new int[N];
    MPI_Status status;
    int ncp=0;

    double dmin = DMAX;
    while (1) {
      MPI_Send(&dmin,1,MPI_DOUBLE,0,0,MPI_COMM_WORLD);
      MPI_Send(pathmin,N,MPI_INT,0,0,MPI_COMM_WORLD);

      MPI_Recv(path,N,MPI_INT,0,MPI_ANY_TAG,
	       MPI_COMM_WORLD,&status);
#ifdef DBG
      printf("[%d] Receives partial path: ",myrank);
      print_path(path,N);
#endif
      if (path[0]==-1) break;
      dmin = DMAX;
      while (1) {
	ncp++;
	double D = dist(path,N,d);
	if (D<dmin) { 
	  dmin = D;
	  memcpy(pathmin,path,N*sizeof(int));
	}
	if(!next_path(path,N,Np)) break;
      }
#ifdef DBG
      printf("[%d] total computed %d\n",myrank,ncp);
#endif
    }

    delete[] path;
    delete[] pathmin;
  }

  delete[] d;
  delete[] x;
  MPI_Finalize();
  
}
