// $Id: tsp1.cpp,v 1.2 2004/10/10 14:33:14 mstorti Exp $
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
//  Solves the Traveling Salesman Problem by exhaustive search
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
#include <cstdio>
#include <cassert>
#include <cmath>
#include <cfloat>
#include <cstring>

using namespace std;

void print_path(int *path,int Nv) {
  printf("(");
  for (int j=0; j<Nv-1; j++) 
    printf("%d,",path[j]);
  printf("%d)\n",path[Nv-1]);
}

#if 0
int next_path(int *path,int N,int Np=0) {
  int j;
  for (j=0; j<N-Np; j++) if (path[j] != N-1-j) break;
  if (j==N-Np) return 0;
  else {
    if (j>0) {
      for (int k=0; k<N-j; k++) path[k] = path[k+j];
      for (int k=N-j; k<N; k++) path[k] = k;
    }
  }

  int jj=N-1-j;
  // Tries to move vertex `jj' one step to the head
  // Look where is 'jj'
  int k;
  for (k=1; k<N-1; k++) if (path[k] == jj) break;
  assert(k<N);
  // Exchange 'jj' with previous position
  path[k] = path[k-1];
  path[k-1] = jj;
  return 1;
}
#else
int next_path(int *path,int N,int Np=0) { //L-next-path-b
  for (int j=N-1; j>=Np; j--) {
    if (path[0]==j) {
      for (int k=0; k<j; k++) path[k] = path[k+1];
      path[j] = j;
    } else {
      int k;
      for (k=0; k<N-1; k++) 
	if (path[k]==j) break;
      path[k] = path[k-1];
      path[k-1] = j;
      return 1;
    }
  }
  return 0;
}
#endif                                     //L-next-path-e

double dist(int *path, int N, double *d) {
  double D=0.0;
  for (int j=0; j<N; j++) {
    int k = j+1;
    if (j==N-1) k=0;
    int J = path[j];
    int K = path[k];
    D += d[N*J+K];
  }
  return D;
}

int main() {
  int n=2, m=2;
  int N = m*n, Np=2;

  double *x = new double[2*N];
  int vertex=0;
  double hx = (m>1 ? 1.0/double(m-1) : 0.0);
  double hy = (n>1 ? 1.0/double(n-1) : 0.0);
  for (int j=0; j<m; j++) {
    for (int k=0; k<n; k++) {
      x[2*vertex] = j*hx;
      x[2*vertex+1] = k*hy;
      vertex++;
    }
  }

  double *d = new double[N*N];
  for (int j=0; j<N; j++) {
    for (int k=0; k<N; k++) {
      double dx = x[2*k] - x[2*j];
      double dy = x[2*k+1] - x[2*j+1];
      d[N*j+k] = sqrt(dx*dx+dy*dy);
    }
  }

  int *path= new int[N];
  int *ppath= new int[N];
  int *pathmin = new int[N];
  for (int j=0; j<N; j++) ppath[j]=j;

  int npp=0, ncp=0;
  double dmin = DBL_MAX;
  int dbg = 1;
  while(1) {
    npp++;
    if (dbg) {
      printf("partial path:");
      print_path(path,N);
      printf("Derived complete paths:\n");
    }
    memcpy(path,ppath,N*sizeof(int));
    while (1) {
      ncp++;
      double D = dist(path,N,d);
      if (dbg) {
	print_path(path,N);
	printf("dist = %f\n",D);
      }
      if (D<dmin) { 
	dmin = D;
	memcpy(pathmin,path,N*sizeof(int));
      }
      if(!next_path(path,N,Np)) break;
    }
    if(!next_path(ppath,Np)) break;
  }
  printf("%d partial paths, %d complete paths\n",npp,ncp);
  printf("min dist = %f\n",dmin);
  printf("min path = "); print_path(pathmin,N);

  delete[] path;
  delete[] ppath;
  delete[] pathmin;
  delete[] d;
  delete[] x;
}
