// $Id: tsp1d.cpp,v 1.2 2004/10/10 14:33:14 mstorti Exp $
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
//  Generates all complete paths from partial paths
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
#include <cstdio>
#include <cassert>
#include <cmath>
#include <cfloat>
#include <cstring>
#include <cstdlib>
#include <omp.h>

using namespace std;

void print_path(int *path,int Nv) {
  printf("(");
  for (int j=0; j<Nv-1; j++) 
    printf("%d,",path[j]);
  printf("%d)\n",path[Nv-1]);
}

int next_path(int *path,int Nv,int Np=0) {
  for (int j=Nv-1; j>=Np; j--) {
    if (path[0]==j) {
      for (int k=0; k<j; k++) path[k] = path[k+1];
      path[j] = j;
    } else {
      int k;
      for (k=0; k<Nv-1; k++) 
	if (path[k]==j) break;
      path[k] = path[k-1];
      path[k-1] = j;
      return 1;
    }
  }
  return 0;
}

double dist(int *path, int Nv, double *d) {
  double D=0.0;
  for (int j=0; j<Nv; j++) {
    int k = j+1;
    if (j==Nv-1) k=0;
    int J = path[j];
    int K = path[k];
    D += d[Nv*J+K];
  }
  return D;
}

//#define DBG

int main(int argc, char **argv) {
  int 
    Nv=12, 
    Np=8;
  if (argc==3) {
    Nv = atoi(argv[1]);
    Np = atoi(argv[2]);
  }

  int *ppath= new int[Nv];
  for (int j=0; j<Nv; j++) ppath[j]=j;
  double start = omp_get_wtime();
  int  ok=1;
#pragma omp parallel
  {
    int *path= new int[Nv], 
      tid = omp_get_thread_num(),
      lok;
    double nphere=0;
    while(1) {
#pragma omp critical 
      { 
        lok = ok;
        if (ok) {
          memcpy(path,ppath,Nv*sizeof(int)); 
          ok = next_path(ppath,Np);
        }
      }
#ifdef DBG
      printf("[%d] partial path: ",tid);
      print_path(path,Np);
#endif
      if (!lok) break;
      while (1) {
#ifdef DBG
        printf("[%d] complete path: ",tid);
        print_path(path,Nv);
#endif
        if (!next_path(path,Nv,Np)) break;
      }

    }
    delete[] path;
  }
  delete[] ppath;
  
  double nppath=1, npath=1;
  for (int j=1; j<Np; j++)
    nppath *= j;
  for (int j=1; j<Nv; j++)
    npath *= j;

  double 
    elapsed = omp_get_wtime()-start,
    rate = elapsed/npath*1e9;
  printf("Nv %d, Np %d, total %.3g paths, %.3g ppaths\n"
         "elapsed %.3f, rate %.3f[secs/Gpath]\n",
         Nv, Np, npath, nppath, elapsed,rate);
}
