//$Id$
#include <mpi.h>
#include <cstdio>
#include <cassert>
#include <vector>
#include <algorithm>

using namespace std;

struct mypair {
  int a,final_pos;
};

int comp(mypair a,mypair b) {
  return a.final_pos<b.final_pos;
}

vector<int> Bsol;
int my_displ;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void vecperm(vector<int> &A, vector<int> &P,vector<int> &B,
             MPI_Comm comm) {
  int myrank, size;
  MPI_Comm_rank(comm,&myrank);
  MPI_Comm_size(comm,&size);
  unsigned int Nhere = A.size();
  assert(P.size()==Nhere);
  vector<mypair> sbuff(Nhere), rbuff(Nhere);
  for (unsigned int j=0; j<Nhere; j++) {
    sbuff[j].a = A[j];
    sbuff[j].final_pos = P[j];
  }
  sort(sbuff.begin(),sbuff.end(),comp);
  vector<int> Nh(size),
    sdispls(size),scnts(size),
    rdispls(size),rcnts(size);
    
  MPI_Allgather(&Nhere,1,MPI_INT,
                &Nh[0],1,MPI_INT,comm);
  vector<int> adispls(size+1);
  adispls[0] = 0;
  for (int j=0; j<size; j++)
    adispls[j+1] = adispls[j] + Nh[j];


#if 0
  if (!myrank) {
    for (int j=0; j<size; j++) {
      printf("[%d] j %d, Nh %d, adispls %d\n",
             myrank,j,Nh[j],adispls[j]);
    }
  }
#endif

  // Scan buffer to send looking the divisions for
  // the sector for each processor
  int drank = 0;
  for (int j=0; j<Nhere; j++) {
    if (sbuff[j].final_pos >= adispls[drank])
      sdispls[drank++] = j;
  }

#if 0
  for (int j=0; j<Nhere; j++)
    printf("[%d] j %d, A %d, P %d\n",myrank,j,sbuff[j].a,sbuff[j].final_pos);
  if (!myrank)
    for (int j=0; j<size; j++)
      printf("[%d] sdispls %d\n",j,sdispls[j]);
#endif

  for (int j=0; j<size-1; j++) 
    scnts[j] = sdispls[j+1]-sdispls[j];
  scnts[size-1] = Nhere-sdispls[size-1];

  for (int j=0; j<size; j++) {
    scnts[j] *= sizeof(mypair);
    sdispls[j] *= sizeof(mypair);
  }

#if 0
  printf("[%d] scnts ",myrank);
  for (int j=0; j<size; j++) 
    printf("%d ",scnts[j]);
  printf("\n",myrank);
#endif
  
  MPI_Alltoall(&scnts[0],1,MPI_INT,
	       &rcnts[0],1,MPI_INT,
               comm);
  rdispls[0] = 0;
  for (int j=0; j<size; j++)
    rdispls[j+1] = rdispls[j] + rcnts[j];

#if 0
  printf("[%d] rcnts ",myrank);
  for (int j=0; j<size; j++) 
    printf("%d ",rcnts[j]);
  printf("\n",myrank);
#endif

  MPI_Alltoallv(&sbuff[0],&scnts[0],&sdispls[0],MPI_CHAR,
                &rbuff[0],&rcnts[0],&rdispls[0],MPI_CHAR,
		comm);
  sort(rbuff.begin(),rbuff.end(),comp);

#if 0
  if (!myrank) 
    for (int j=0; j<Nhere; j++) 
      printf("j %d, A %d, finpos %d\n",
             j,rbuff[j].a,rbuff[j].final_pos);
#endif

  B.resize(Nhere);
  for (int j=0; j<Nhere; j++) {
    assert(rbuff[j].final_pos == adispls[myrank]+j);
    B[j] = rbuff[j].a;
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void vecperm_easy(vector<int> &A,vector<int> &P,
                  vector<int> &B, MPI_Comm comm) {
  int myrank, size;
  MPI_Comm_rank(comm,&myrank);
  MPI_Comm_size(comm,&size);
  unsigned int Nhere = A.size();
  assert(P.size()==Nhere);
  vector<int> Nh(size),rdispls(size);
  MPI_Allgather(&Nhere,1,MPI_INT,
                &Nh[0],1,MPI_INT,comm);
  int N = 0;
  rdispls[0] = 0;
  for (int j=0; j<size; j++) {
    N += Nh[j];
    if (j>0) 
      rdispls[j] = rdispls[j-1]+Nh[j-1];
  }

  vector<int> Ag(N),Pg(N),Bg(N);
  MPI_Allgatherv(&A[0],Nhere,MPI_INT,
                 &Ag[0],&Nh[0],&rdispls[0],MPI_INT,comm);
  MPI_Allgatherv(&P[0],Nhere,MPI_INT,
                 &Pg[0],&Nh[0],&rdispls[0],MPI_INT,comm);

  for (int j=0; j<N; j++)
    Bg[Pg[j]] = Ag[j];

  B.resize(Nhere);
  for (int j=0; j<Nhere; j++)
    B[j] = Bg[rdispls[myrank]+j];
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void vecperm_init(vector<int> &A, vector<int> &P,
                  int N,MPI_Comm comm) {
  int myrank, size;
  MPI_Comm_rank(comm,&myrank);
  MPI_Comm_size(comm,&size);
  vector<int> Nh(size);
  for (int j=0; j<size; j++) 
    Nh[j] = N/size+(j<N%size);

  vector<mypair> AP(N);
  Bsol.resize(N);
  if (!myrank) {
    for (int j=0; j<N; j++) {
      AP[j].a = rand()%1000;
      AP[j].final_pos = j;
    }
    for (int j=0; j<N; j++)
      Bsol[j] = AP[j].a;
    random_shuffle(AP.begin(),AP.end());

  }
  MPI_Bcast(&AP[0],N*sizeof(mypair),MPI_CHAR,0,MPI_COMM_WORLD);
  MPI_Bcast(&Bsol[0],N,MPI_INT,0,MPI_COMM_WORLD);

  int Nhere = Nh[myrank];
  A.resize(Nhere);
  P.resize(Nhere);
  int base=0;
  for (int j=0; j<myrank; j++) 
    base += Nh[j];
  my_displ = base;

  for (int j=0; j<Nhere; j++) {
    A[j] = AP[base+j].a;
    P[j] = AP[base+j].final_pos;
  }  
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void vecperm_init(int **A, int **P,
                  int N,int *Nherep,MPI_Comm comm) {
  vector<int> Av,Pv;
  vecperm_init(Av,Pv,N,MPI_COMM_WORLD);
  *Nherep = Av.size();
  assert(Av.size()==Pv.size());
  *A = (int *)malloc((*Nherep) * sizeof(int));
  *P = (int *)malloc((*Nherep) * sizeof(int));
  for (int j=0; j<*Nherep; j++) {
    int kk = Av[j]; 
    (*A)[j] = kk;
    (*P)[j] = Pv[j];
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main(int argc, char **argv) {
  MPI_Init(&argc,&argv);

  int myrank, size;
  MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
  MPI_Comm_size(MPI_COMM_WORLD,&size);

  int N = 20;
  vector<int> A, P, B;
#if 0
  vecperm_init(A,P,N,MPI_COMM_WORLD);
  int Nhere = A.size();
#else
  int *Ap, *Pp, Nhere;
  vecperm_init(&Ap,&Pp,N,&Nhere,MPI_COMM_WORLD);
  A.resize(Nhere);
  P.resize(Nhere);
  for (int j=0; j<Nhere; j++) {
    A[j] = Ap[j];
    P[j] = Pp[j];
  }
  free(Ap);
  free(Pp);
#endif

  vecperm_easy(A,P,B,MPI_COMM_WORLD);
  vecperm(A,P,B,MPI_COMM_WORLD);
  
#if 1
  for (int j=0; j<Nhere; j++) {
    printf("[%d] j %d, A %d, P %d, B[%d], Bsol %d\n",
           myrank,j,A[j],P[j],B[j],Bsol[my_displ+j]);
  }
#endif
  MPI_Finalize();
}
